import getpass
import hashlib
import os
import sys

import yaml
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7


class Keystore:
    def __init__(self, config):
        self.keylist_file = config.path / "keys"
        p = getpass.getpass("Enter encryption phrase for key database: ")
        sha = hashlib.sha256(p.encode("ascii"))
        self.nonce_file = config.path / "nonce"
        if not self.nonce_file.exists():
            self.nonce = os.urandom(16)
            with self.nonce_file.open("wb") as f:
                f.write(self.nonce)
        else:
            with self.nonce_file.open("rb") as f:
                self.nonce = f.read()
        self.keylist_cipher = Cipher(AES(sha.digest()), CBC(self.nonce))

        if not self.keylist_file.exists():
            self.keys = {}
        else:
            self.parse_keylist()

    def parse_keylist(self):
        try:
            with self.keylist_file.open("rb") as f:
                data = f.read()
            dec = self.keylist_cipher.decryptor()
            yd = dec.update(data) + dec.finalize()
            unpadder = PKCS7(256).unpadder()
            yd = unpadder.update(yd) + unpadder.finalize()
            self.keys = yaml.safe_load(yd.decode("UTF-8"))
        except ValueError:
            print("Incorrect password.")
            sys.exit(1)

    def write_keylist(self):
        enc = self.keylist_cipher.encryptor()
        yd = yaml.dump(self.keys).encode("UTF-8")
        padder = PKCS7(256).padder()
        padded = padder.update(yd) + padder.finalize()
        data = enc.update(padded) + enc.finalize()
        with self.keylist_file.open("wb") as f:
            f.write(data)

    def __getitem__(self, item):
        return self.keys[item]

    def set_key(self, name, access, secret):
        self.keys[name] = {"access": access, "secret": secret}
        self.write_keylist()

    def delete_key(self, name):
        del self.keys[name]
        self.write_keylist()



import re
from pathlib import Path

from .common import Common, DefaultAnchor, DefaultBorder, DefaultDimension
from .info import HotkeyDisplay
from .termui.alignment import CenterAnchor, Dimension, TopRightAnchor
from .termui.control import Border
from .termui.dialog import DialogControl, DialogFieldText
from .termui.list_control import ListControl, ListEntry


class SSHList(ListControl):
    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            **kwargs,
        )
        l.border = DefaultBorder("ssh", "SSH Keys", None)
        return [l, l.hotkey_display]

    @classmethod
    def selector(cls, cb, **kwargs):
        return cls.opener(**{"selector_cb": cb, **kwargs})

    def __init__(
        self, parent, alignment, dimensions, selector_cb=None, *args, **kwargs
    ):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("ssh_key_list_generic", "generic"),
            selection_color=Common.color("ssh_key_list_selection", "selection"),
            title_color=Common.color("ssh_key_list_heading", "column_title"),
            *args,
            **kwargs,
        )
        self.selector_cb = selector_cb
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.add_hotkey("d", self.set_default_ssh_key, "Set as default")
        self.add_hotkey("e", self.set_default_ssh_username, "Set default username")
        if selector_cb is not None:
            self.add_hotkey("KEY_ENTER", self.select_and_close, "Select and close")
        else:
            self.add_hotkey("KEY_ENTER", self.select_ssh_key, "Select ssh key")
        self.add_column("usage frequency", 12)
        self.add_column("has public key", 12)
        self.add_column("default username", 20)
        self.add_column("default", 8)
        self.reload()

    def reload(self, move=True):
        self.entries = []
        idx = 0
        defa = 0
        for ssh_key in sorted(self.list_ssh_keys()):
            if ssh_key == Common.Configuration["default_ssh_key"]:
                defa = idx
                d = "✓"
            else:
                d = " "
            pubkey = Path.home() / ".ssh" / "{0}.pub".format(ssh_key)
            has_pubkey = "✓" if pubkey.exists() else " "
            du = (
                Common.Configuration["default_ssh_usernames"][ssh_key]
                if ssh_key in Common.Configuration["default_ssh_usernames"]
                else ""
            )
            self.add_entry(
                ListEntry(
                    ssh_key,
                    **{
                        "usage frequency": 0,
                        "has public key": has_pubkey,
                        "default username": du,
                        "default": d,
                    },
                )
            )
            idx += 1
        if move:
            self.selected = defa

    def list_ssh_keys(self):
        ssh_dir = Path.home() / ".ssh"
        ret = []
        for child in ssh_dir.iterdir():
            fn = child.name
            if re.match(r"^((id_[^.]+?)|(.*?\.pem))$", fn) is not None:
                ret.append(fn)
        return ret

    def set_default_ssh_key(self, _):
        if self.selection is not None:
            Common.Configuration["default_ssh_key"] = self.selection.name
            Common.Configuration.write_config()
            for entry in self.entries:
                if entry.name != Common.Configuration["default_ssh_key"]:
                    entry.columns["default"] = " "
                else:
                    entry.columns["default"] = "✓"

    def select_ssh_key(self, _):
        if self.selection is not None:
            Common.Session.ssh_key = self.selection.name

    def select_and_close(self, _):
        if self.selection is not None and self.selector_cb is not None:
            self.selector_cb(self.selection["name"])
            Common.Session.pop_frame()

    def set_default_ssh_username(self, _):
        SSHDefaultUsernameDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            key_name=self.selection["name"],
            caller=self,
            weight=-500,
        )


class SSHDefaultUsernameDialog(DialogControl):
    def __init__(
        self, parent, alignment, dimensions, key_name="", caller=None, *args, **kwargs
    ):
        self.key_name = key_name
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        kwargs["border"] = Border(
            Common.border("ssh", "default"),
            Common.color("ssh_modal_dialog_border", "modal_dialog_border"),
            "Default SSH username",
            Common.color("ssh_modal_dialog_border_title", "modal_dialog_border_title"),
            self.key_name,
            Common.color(
                "ssh_modal_dialog_border_title_info", "modal_dialog_border_title_info"
            ),
        )
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        def_text = (
            Common.Configuration["default_ssh_usernames"][self.key_name]
            if self.key_name in Common.Configuration["default_ssh_usernames"]
            else ""
        )
        self.username_textfield = DialogFieldText(
            "SSH username",
            text=def_text,
            color=Common.color(
                "ec2_ssh_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_label", "modal_dialog_textfield_label"
            ),
            label_min=16,
        )
        self.add_field(self.username_textfield)
        self.highlighted = 1 if def_text != "" else 0
        self.caller = caller

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(inkey)

    def accept_and_close(self):
        Common.Configuration["default_ssh_usernames"][
            self.key_name
        ] = self.username_textfield.text
        Common.Configuration.write_config()
        Common.Session.ssh_key = Common.Session.ssh_key  # don't ask
        self.caller.reload(move=False)
        self.close()

    def close(self):
        self.parent.remove_block(self)



from .base_control import Describer, ResourceLister
from .common import Common


class TargetGroupResourceLister(ResourceLister):
    prefix = "target_group_list"
    title = "Target Groups"
    command_palette = ["tg", "targetgroup"]

    def title_info(self):
        if self.rule is not None:
            return self.rule["arn"]
        return None

    def __init__(self, *args, **kwargs):
        self.resource_key = "elbv2"
        self.list_method = "describe_target_groups"
        self.item_path = ".TargetGroups"
        if "rule_entry" in kwargs:
            self.rule = kwargs["rule_entry"]
            arns = []
            raw = self.rule.controller_data
            try:
                arns = [
                    i["TargetGroupArn"]
                    for i in raw["Actions"][-1]["ForwardConfig"]["TargetGroups"]
                ]
                self.list_kwargs = {"TargetGroupArns": arns}
            except KeyError:
                self.rule = None
                Common.Session.set_message(
                    "Rule is not configured for forwarding.",
                    Common.color("message_error"),
                )
        else:
            self.rule = None
        self.column_paths = {
            "name": ".TargetGroupName",
            "protocol": ".Protocol",
            "port": ".Port",
            "target type": ".TargetType",
        }
        self.imported_column_sizes = {
            "name": 30,
            "protocol": 10,
            "port": 10,
            "target type": 10,
        }
        self.hidden_columns = {
            "arn": ".TargetGroupArn",
        }
        self.describe_command = TargetGroupDescriber.opener
        # self.open_command = ListenerActionResourceLister.opener
        # self.open_selection_arg = 'listener'

        self.imported_column_order = ["name", "protocol", "port", "target type"]
        self.sort_column = "name"
        self.primary_key = "arn"
        super().__init__(*args, **kwargs)


class TargetGroupDescriber(Describer):
    prefix = "tg_browser"
    title = "Target Group"

    def populate_entry(self, *args, entry, **kwargs):
        super().populate_entry(*args, entry=entry, **kwargs)
        try:
            self.name = entry["name"]
        except KeyError:
            self.name = self.entry_id

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="arn", **kwargs
    ):
        self.resource_key = "elbv2"
        self.describe_method = "describe_target_groups"
        self.describe_kwarg_name = "TargetGroupArns"
        self.describe_kwarg_is_list = True
        self.object_path = ".TargetGroups[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )

    def title_info(self):
        return self.name



from .base_control import Describer, ResourceLister


class LBResourceLister(ResourceLister):
    prefix = "lb_list"
    title = "Load Balancers"
    command_palette = ["lb", "elbv2", "loadbalancing"]

    def __init__(self, *args, **kwargs):
        from .resource_listener import ListenerResourceLister

        self.resource_key = "elbv2"
        self.list_method = "describe_load_balancers"
        self.item_path = ".LoadBalancers"
        self.column_paths = {
            "name": ".LoadBalancerName",
            "type": ".Type",
            "scheme": self.determine_scheme,
            "hostname": ".DNSName",
        }
        self.imported_column_sizes = {
            "name": 30,
            "type": 15,
            "scheme": 8,
            "hostname": 120,
        }
        self.hidden_columns = {
            "arn": ".LoadBalancerArn",
        }
        self.describe_command = LBDescriber.opener
        self.open_command = ListenerResourceLister.opener
        self.open_selection_arg = "lb"

        self.imported_column_order = ["name", "type", "scheme", "hostname"]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)

    def determine_scheme(self, lb, *args):
        if lb["Scheme"] == "internet-facing":
            return "public"
        return "private"


class LBDescriber(Describer):
    prefix = "lb_browser"
    title = "Load Balancer"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elbv2"
        self.describe_method = "describe_load_balancers"
        self.describe_kwarg_name = "Names"
        self.describe_kwarg_is_list = True
        self.object_path = ".LoadBalancers[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .base_control import (
    DeleteResourceDialog,
    Describer,
    ResourceLister,
    SingleRelationLister,
    SingleSelectorDialog,
)
from .common import Common
from .termui.alignment import CenterAnchor, Dimension
from .termui.dialog import DialogFieldCheckbox
from .termui.ui import ControlCodes


class EBApplicationResourceLister(ResourceLister):
    prefix = "eb_application_list"
    title = "Beanstalk Applications"
    command_palette = ["ebapplication", "ebapp", "elasticbeanstalkapplication"]

    def delete_application(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="application",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
            can_force=True,
        )

    def do_delete(self, force, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "ApplicationName": self.selection["name"],
            "TerminateEnvByForce": force,
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "delete_application",
            api_kwargs,
            "Delete application",
            "Elastic Beanstalk",
            subcategory="Application",
            success_template="Deleting application {0}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def __init__(self, *args, **kwargs):
        self.resource_key = "elasticbeanstalk"
        self.list_method = "describe_applications"
        self.item_path = ".Applications"
        self.column_paths = {
            "name": ".ApplicationName",
            "description": ".Description",
        }
        self.imported_column_sizes = {
            "name": 20,
            "description": 50,
        }
        self.hidden_columns = {
            "arn": ".ApplicationArn",
        }
        self.describe_command = EBApplicationDescriber.opener
        self.open_command = EBEnvironmentLister.opener
        self.open_selection_arg = "app"

        self.additional_commands = {
            "v": {
                "command": EBApplicationVersionResourceLister.opener,
                "selection_arg": "application",
                "tooltip": "View Versions",
            }
        }

        self.imported_column_order = ["name", "description"]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)


class EBApplicationDescriber(Describer):
    prefix = "eb_application_browser"
    title = "Beanstalk Application"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_applications"
        self.describe_kwarg_name = "ApplicationNames"
        self.describe_kwarg_is_list = True
        self.object_path = ".Applications[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )
        self.add_hotkey(ControlCodes.D, self.delete_application, "Delete application")


class EBApplicationVersionResourceLister(ResourceLister):
    prefix = "eb_application_version_list"
    title = "Beanstalk Application Versions"
    command_palette = [
        "ebapplicationversion",
        "ebappversion",
        "elasticbeanstalkapplicationversion",
    ]

    def delete_application_version(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="application version",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
            can_force=True,
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "ApplicationName": self.selection["application"],
            "VersionLabel": self.selection["version"],
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "delete_application_version",
            api_kwargs,
            "Delete appversion",
            "Elastic Beanstalk",
            subcategory="Application",
            success_template="Deleting application version {0}",
            resource=self.selection["version"],
        )
        self.refresh_data()

    def __init__(self, *args, application=None, **kwargs):
        self.application = application
        self.resource_key = "elasticbeanstalk"
        self.list_method = "describe_application_versions"
        if application is not None:
            self.list_kwargs = {"ApplicationName": self.application["name"]}
        self.item_path = ".ApplicationVersions"
        self.column_paths = {
            "application": ".ApplicationName",
            "version": ".VersionLabel",
            "status": ".Status",
            "description": ".Description",
        }
        self.imported_column_sizes = {
            "name": 20,
            "version": 20,
            "status": 10,
            "description": 50,
        }
        self.hidden_columns = {
            "name": ".VersionLabel",
        }
        self.describe_command = EBApplicationVersionDescriber.opener

        self.imported_column_order = ["name", "version", "status", "description"]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)


class EBApplicationVersionDescriber(Describer):
    prefix = "eb_application_version_browser"
    title = "Beanstalk Application Version"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_application_versions"
        self.describe_kwarg_name = "VersionLabels"
        self.describe_kwarg_is_list = True
        self.describe_kwargs = {"ApplicationName": entry["application"]}
        self.object_path = ".Applications[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )
        self.add_hotkey(
            ControlCodes.D, self.delete_application_version, "Delete app version"
        )


class EBEnvironmentLister(ResourceLister):
    prefix = "eb_environment_list"
    title = "Beanstalk Environments"
    command_palette = ["ebenvironment", "ebenv", "elasticbeanstalkenvironment"]

    def delete_environment(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="environment",
            resource_identifier=self.selection["name"],
            callback=self.do_delete,
            action_name="Delete",
            can_force=True,
            extra_fields={
                "terminate_resources_field": DialogFieldCheckbox(
                    "Terminate resources",
                    checked=False,
                    color=Common.color("modal_dialog_textfield_label"),
                    selected_color=Common.color("modal_dialog_textfield_selected"),
                )
            },
        )

    def do_delete(self, force, terminate_resources_field, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "EnvironmentName": self.selection["name"],
            "ForceTerminate": force,
            "TerminateResources": terminate_resources_field.checked,
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "terminate_environment",
            api_kwargs,
            "Terminate environment",
            "Elastic Beanstalk",
            subcategory="Environment",
            success_template="Terminating environment {0}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def delete_environment_config(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="environment configuration",
            resource_identifier=self.selection["name"],
            callback=self.do_delete_config,
            action_name="Delete",
        )

    def do_delete_config(self, force, terminate_resources_field, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "EnvironmentName": self.selection["name"],
            "ApplicationName": self.selection["application"],
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "delete_environment_configuration",
            api_kwargs,
            "Delete envconfig",
            "Elastic Beanstalk",
            subcategory="Environment",
            success_template="Deleting environment {0} configuration",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def rebuild_environment(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="environment",
            resource_identifier=self.selection["name"],
            callback=self.do_rebuild,
            action_name="Rebuild",
        )

    def do_rebuild(self, **kwargs):
        if self.selection is None:
            return
        api_kwargs = {
            "EnvironmentName": self.selection["name"],
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "rebuild_environment",
            api_kwargs,
            "Rebuild environment",
            "Elastic Beanstalk",
            subcategory="Environment",
            success_template="Rebuilding environment {0}",
            resource=self.selection["name"],
        )
        self.refresh_data()

    def swap_cnames(self, _):
        if self.selection is None:
            return
        SingleSelectorDialog(
            self.parent,
            "Swap CNames of environment '{0}' with other environment".format(
                self.selection["name"]
            ),
            "environment",
            "swap",
            EBEnvironmentLister,
            self.do_swap,
            caller=self,
        )

    def do_swap(self, other_env):
        if self.selection is None:
            return
        if other_env == self.selection["name"]:
            Common.Session.error(
                "Source and destination environments are the same.",
                "Swap Cnames",
                "Elastic Beanstalk",
                resource=self.selection["name"],
            )
            return
        api_kwargs = {
            "SourceEnvironmentName": self.selection["name"],
            "DestinationEnvironmentName": other_env,
        }
        Common.generic_api_call(
            "elasticbeanstalk",
            "swap_environment_cnames",
            api_kwargs,
            "Swap Cnames",
            "Elastic Beanstalk",
            subcategory="Environment",
            success_template="Swapping cnames for environments {0}",
            resource="{0}, {1}".format(self.selection["name"], other_env),
        )
        self.refresh_data()

    def __init__(self, *args, app=None, **kwargs):
        self.resource_key = "elasticbeanstalk"
        self.list_method = "describe_environments"
        self.item_path = ".Environments"
        self.app = app
        if app is not None:
            self.list_kwargs = {"ApplicationName": self.app}
        self.column_paths = {
            "name": ".EnvironmentName",
            "id": ".EnvironmentId",
            "application": ".ApplicationName",
            "status": ".Status",
            "health": ".Health",
            "tier": ".Tier.Name",
        }
        self.imported_column_sizes = {
            "name": 20,
            "id": 20,
            "application": 20,
            "status": 15,
            "health": 7,
            "tier": 10,
        }
        self.hidden_columns = {
            "arn": ".EnvironmentArn",
        }
        self.describe_command = EBEnvironmentDescriber.opener
        # self.open_command = EBEnvironmentLister.opener
        # self.open_selection_arg = 'lb'

        self.imported_column_order = [
            "name",
            "id",
            "tier",
            "status",
            "health",
            "application",
        ]
        self.sort_column = "name"
        self.primary_key = "name"
        self.additional_commands = {
            "v": {
                "command": EBEnvironmentResourceLister.opener,
                "selection_arg": "environment",
                "tooltip": "View env resources",
            },
            "i": {
                "command": EBInstanceHealthLister.opener,
                "selection_arg": "environment",
                "tooltip": "View instance healths",
            },
            "h": {
                "command": EBEnvironmentHealthDescriber.opener,
                "selection_arg": "entry",
                "tooltip": "View env health",
            },
        }
        super().__init__(*args, **kwargs)
        if self.selector_cb is None:
            self.add_hotkey(
                ControlCodes.B, self.rebuild_environment, "Rebuild environment"
            )
            self.add_hotkey(ControlCodes.S, self.swap_cnames, "Swap CNames")
            self.add_hotkey(
                ControlCodes.D, self.delete_environment, "Terminate environment"
            )
            self.add_hotkey(
                ControlCodes.E, self.delete_environment, "Delete env config"
            )

    def title_info(self):
        return self.app


class EBEnvironmentDescriber(Describer):
    prefix = "eb_environment_browser"
    title = "Beanstalk Environment"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_environments"
        self.describe_kwarg_name = "EnvironmentNames"
        self.describe_kwarg_is_list = True
        self.object_path = ".Environments[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class EBEnvironmentHealthDescriber(Describer):
    prefix = "eb_environment_health_browser"
    title = "Beanstalk Environment Health"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_environment_health"
        self.describe_kwarg_name = "EnvironmentName"
        self.describe_kwarg_is_list = True
        self.object_path = "."
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class EBEnvironmentResourceLister(SingleRelationLister):
    prefix = "eb_environment_resources"
    title = "Beanstalk Environment resources"

    def title_info(self):
        return self.environment["name"]

    def __init__(self, *args, environment=None, **kwargs):
        from .resource_asg import ASGDescriber
        from .resource_ec2 import EC2Describer
        from .resource_lb import LBDescriber
        from .resource_lc import LCDescriber
        from .resource_sqs import SQSDescriber

        self.resource_key = "elasticbeanstalk"
        self.environment = environment
        self.describe_method = "describe_environment_resources"
        self.describe_kwargs = {"EnvironmentName": self.environment["name"]}
        self.object_path = ".EnvironmentResources"
        self.resource_descriptors = [
            {
                "base_path": "[.AutoscalingGroups[].Name]",
                "type": "Auto Scaling Group",
                "describer": ASGDescriber.opener,
            },
            {
                "base_path": "[.Instances[].Id]",
                "type": "EC2 Instance",
                "describer": EC2Describer.opener,
            },
            {
                "base_path": "[.LaunchConfigurations[].Name]",
                "type": "Launch Configuration",
                "describer": LCDescriber.opener,
            },
            {
                "base_path": "[.LaunchTemplates[].Name]",
                "type": "Launch Templates",
            },
            {
                "base_path": "[.LoadBalancers[].Name]",
                "type": "Load Balancer",
                "describer": LBDescriber.opener,
            },
            {
                "base_path": "[.Triggers[].Name]",
                "type": "Trigger",
            },
            {
                "base_path": "[.Queues[].URL]",
                "type": "SQS Queue",
                "describer": SQSDescriber.opener,
            },
        ]
        super().__init__(*args, **kwargs)


class EBPlatformBranchLister(ResourceLister):
    prefix = "eb_platform_version_list"
    title = "Beanstalk Platform Branches"
    command_palette = [
        "platformbranch",
        "ebplatformbranch",
        "elasticbeanstalkplatformbranch",
    ]

    def __init__(self, *args, **kwargs):
        self.resource_key = "elasticbeanstalk"
        self.list_method = "list_platform_branches"
        self.item_path = ".PlatformBranchSummaryList"
        self.column_paths = {
            "platform": ".PlatformName",
            "branch": ".BranchName",
            "lifecycle": ".LifecycleState",
            "order": ".BranchOrder",
        }
        self.imported_column_sizes = {
            "platform": 20,
            "branch": 20,
            "lifecycle": 20,
            "order": 10,
        }
        self.hidden_columns = {
            "name": ".BranchName",
        }
        self.describe_command = EBPlatformVersionDescriber.opener

        self.imported_column_order = [
            "branch",
            "platform",
            "lifecycle",
            "order",
        ]
        self.sort_column = "branch"
        self.primary_key = "name"
        self.additional_commands = {
            "v": {
                "command": EBPlatformVersionLister.opener,
                "selection_arg": "branch",
                "tooltip": "View Versions",
            }
        }
        super().__init__(*args, **kwargs)


class EBPlatformBranchDescriber(Describer):
    prefix = "eb_platform_branch_browser"
    title = "Beanstalk Platform Branch"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="arn", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_platform_version"
        self.describe_kwarg_name = "PlatformArn"
        self.object_path = ".PlatformDescription"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class EBPlatformVersionLister(ResourceLister):
    prefix = "eb_platform_version_list"
    title = "Beanstalk Platform Versions"
    command_palette = [
        "platformversion",
        "ebplatformversion",
        "elasticbeanstalkplatformversion",
    ]

    def title_info(self):
        if self.branch is not None:
            return "Branch: {0}".format(self.branch["name"])
        return None

    def __init__(self, *args, branch=None, **kwargs):
        self.branch = branch
        self.resource_key = "elasticbeanstalk"
        self.list_method = "list_platform_versions"
        if self.branch is not None:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Type": "PlatformBranchName",
                        "Operator": "=",
                        "Values": [self.branch["name"]],
                    }
                ]
            }
        self.item_path = ".PlatformSummaryList"
        self.column_paths = {
            "owner": ".PlatformOwner",
            "os": ".OperatingSystemName",
            "os version": ".OperatingSystemVersion",
            "status": ".PlatformStatus",
            "category": ".PlatformCategory",
            "version": ".PlatformVersion",
            "lifecycle": ".PlatformLifecycleState",
        }
        self.imported_column_sizes = {
            "owner": 10,
            "os": 20,
            "os version": 10,
            "status": 10,
            "category": 10,
            "version": 10,
            "lifecycle": 10,
        }
        self.hidden_columns = {
            "name": ".PlatformArn",
            "arn": ".PlatformArn",
        }
        self.describe_command = EBPlatformVersionDescriber.opener

        self.imported_column_order = [
            "owner",
            "os",
            "os version",
            "category",
            "version",
            "status",
            "lifecycle",
        ]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)


class EBPlatformVersionDescriber(Describer):
    prefix = "eb_platform_version_browser"
    title = "Beanstalk Platform Version"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="arn", **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_platform_version"
        self.describe_kwarg_name = "PlatformArn"
        self.object_path = ".PlatformDescription"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class EBInstanceHealthLister(ResourceLister):
    prefix = "eb_instance_health_list"
    title = "Beanstalk Instance Health"

    def __init__(self, *args, environment=None, **kwargs):
        from .resource_ec2 import EC2Describer

        self.environment = environment
        self.resource_key = "elasticbeanstalk"
        self.list_method = "describe_instances_health"
        self.list_kwargs = {"EnvironmentName": environment["name"]}
        self.item_path = ".InstanceHealthList"
        self.column_paths = {
            "instance id": ".InstanceId",
            "status": ".HealthStatus",
            "color": ".Color",
            "az": ".AvailabilityZone",
            "type": ".InstanceType",
        }
        self.imported_column_sizes = {
            "instance id": 20,
            "az": 20,
            "type": 20,
            "color": 20,
            "status": 20,
        }
        self.hidden_columns = {
            "name": ".InstanceId",
        }
        self.describe_command = EBInstanceHealthDescriber.opener
        self.additional_commands = {
            "v": {
                "command": EC2Describer.opener,
                "selection_arg": "entry",
                "tooltip": "Describe EC2 Instance",
            }
        }

        self.imported_column_order = [
            "instance id",
            "az",
            "type",
            "color",
            "status",
        ]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)


class EBInstanceHealthDescriber(Describer):
    prefix = "eb_instance_health_browser"
    title = "Beanstalk Instance"

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="instance id",
        caller=None,
        **kwargs
    ):
        self.resource_key = "elasticbeanstalk"
        self.describe_method = "describe_instances_health"
        self.describe_kwargs = {"EnvironmentName": caller.environment["name"]}
        self.object_path = ".InstanceList[] | select(.InstanceId=={0})".format(
            entry[entry_key]
        )
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .base_control import Describer, ResourceLister


class LCResourceLister(ResourceLister):
    prefix = "lc_list"
    title = "Launch Configurations"
    command_palette = ["lc", "launchconfiguration"]

    def __init__(self, *args, **kwargs):
        from .resource_asg import ASGResourceLister

        self.resource_key = "autoscaling"
        self.list_method = "describe_launch_configurations"
        self.item_path = ".LaunchConfigurations"
        self.column_paths = {
            "name": ".LaunchConfigurationName",
            "image id": ".ImageId",
            "instance type": ".InstanceType",
        }
        self.imported_column_sizes = {
            "name": 30,
            "image id": 20,
            "instance type": 20,
        }
        self.describe_command = LCDescriber.opener
        self.open_command = ASGResourceLister.opener
        self.open_selection_arg = "lc"

        self.imported_column_order = ["name", "image id", "instance type"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)


class LCDescriber(Describer):
    prefix = "lc_browser"
    title = "Launch Configuration"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "autoscaling"
        self.describe_method = "describe_launch_configurations"
        self.describe_kwarg_name = "LaunchConfigurationNames"
        self.describe_kwarg_is_list = True
        self.object_path = ".LaunchConfigurations[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



class ARN:
    def __init__(self, s):
        parts = s.split(":")
        if len(parts) < 6:
            raise ValueError(
                "ARN does not contain enough components for full identification"
            )
        self.partition = parts[1] if parts[1] != "" else "aws"
        if parts[2] == "":
            raise ValueError("ARN is missing service identifier.")
        self.service = parts[2]
        self.region = parts[3]
        self.account_id = parts[4]
        if len(parts) == 6:
            test = parts[5].split("/")
            if len(test) > 1:
                self.resource_type = test[0]
                self.resource_id = "/".join(test[1:])
            else:
                self.resource_id = parts[5]
                self.resource_type = ""
        else:
            self.resource_type = parts[5]
            self.resource_id = ":".join(parts[6:])

    @property
    def resource(self):
        if self.resource_type != "":
            return "{0}/{1}".format(self.resource_type, self.resource_id)
        return self.resource_id

    @property
    def resource_id_first(self):
        return self.resource_id.split("/")[0]

    def __str__(self):
        return "arn:{0}:{1}:{2}:{3}:{4}".format(
            self.partition, self.service, self.region, self.account_id, self.resource
        )



from .base_control import Describer, ResourceLister


class DBSubnetGroupResourceLister(ResourceLister):
    prefix = "db_subnet_group_list"
    title = "DB Subnet Groups"
    command_palette = ["dsg", "dbsubnetgroup"]

    def __init__(self, *args, **kwargs):
        from .resource_subnet import SubnetResourceLister

        self.resource_key = "rds"
        self.list_method = "describe_db_subnet_groups"
        self.item_path = ".DBSubnetGroups"
        self.column_paths = {
            "name": ".DBSubnetGroupName",
            "vpc": ".VpcId",
            "status": ".SubnetGroupStatus",
        }
        self.hidden_columns = {
            "arn": ".DBSubnetGroupArn",
        }
        self.imported_column_sizes = {
            "name": 30,
            "vpc": 30,
            "status": 30,
        }
        self.describe_command = DBSubnetGroupDescriber.opener
        self.open_command = SubnetResourceLister.opener
        self.open_selection_arg = "db_subnet_group"
        self.primary_key = "name"

        self.imported_column_order = ["name", "vpc", "status"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)


class DBSubnetGroupDescriber(Describer):
    prefix = "subnet_browser"
    title = "Subnet"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "rds"
        self.describe_method = "describe_db_subnet_groups"
        self.describe_kwarg_name = "DBSubnetGroupName"
        self.describe_kwarg_is_list = True
        self.object_path = ".DBSubnetGroups[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



import json

from .base_control import Describer, GenericDescriber, ResourceLister
from .common import Common


class RouteTableResourceLister(ResourceLister):
    prefix = "route_table_list"
    title = "Route Tables"
    command_palette = ["rt", "routetable"]

    def title_info(self):
        if self.subnet is not None:
            return self.subnet["id"]
        return None

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_route_tables"
        self.item_path = ".RouteTables"
        if "subnet" in kwargs:
            self.subnet = kwargs["subnet"]
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "association.subnet-id",
                        "Values": [
                            self.subnet["id"],
                        ],
                    }
                ]
            }
        else:
            self.subnet = None
        self.column_paths = {
            "id": ".RouteTableId",
            "name": self.tag_finder_generator("Name"),
            "vpc": ".VpcId",
            "subnet": self.determine_subnet_association,
        }
        self.imported_column_sizes = {
            "id": 30,
            "name": 30,
            "vpc": 30,
            "subnet": 30,
        }
        self.describe_command = RouteTableDescriber.opener
        self.open_command = RouteResourceLister.opener
        self.open_selection_arg = "route_table"

        self.imported_column_order = ["id", "name", "vpc", "subnet"]
        self.sort_column = "id"
        self.primary_key = "id"
        super().__init__(*args, **kwargs)

    def determine_subnet_association(self, rt, *args):
        if "Associations" not in rt or len(rt["Associations"]) == 0:
            return "<none>"
        if len(rt["Associations"]) > 1:
            return "<multiple>"
        if "SubnetId" in rt["Associations"][0]:
            return rt["Associations"][0]["SubnetId"]
        else:
            return "<VPC default>"


class RouteTableDescriber(Describer):
    prefix = "route_table_browser"
    title = "Route Table"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_route_tables"
        self.describe_kwarg_name = "RouteTableIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".RouteTables[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


# Routes
class RouteResourceLister(ResourceLister):
    prefix = "route_list"
    title = "Routes"
    command_palette = ["route"]

    def title_info(self):
        if self.route_table is not None:
            return self.route_table["id"]
        return None

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_route_tables"
        self.item_path = "[.RouteTables[] as $rt | $rt.Routes[] as $r | $r | .RouteTableId = $rt.RouteTableId]"
        if "route_table" in kwargs:
            self.route_table = kwargs["route_table"]
            self.list_kwargs = {"RouteTableIds": [self.route_table["id"]]}
        else:
            self.route_table = None
        self.column_paths = {
            "route table": ".RouteTableId",
            "gateway type": self.determine_gateway_type,
            "gateway": self.determine_gateway,
            "destination": ".DestinationCidrBlock",
            "state": ".State",
        }
        self.hidden_columns = {
            "name": self.empty,
        }
        self.imported_column_sizes = {
            "route table": 30,
            "gateway type": 20,
            "gateway": 30,
            "destination": 30,
            "state": 10,
        }

        self.imported_column_order = [
            "route table",
            "gateway type",
            "gateway",
            "destination",
            "state",
        ]
        self.sort_column = "route table"
        self.primary_key = None
        super().__init__(*args, **kwargs)

        self.add_hotkey("d", self.generic_describe, "Describe")
        self.add_hotkey("KEY_ENTER", self.generic_describe, "Describe")

    def generic_describe(self, entry):
        if self.selection is not None:
            Common.Session.push_frame(
                GenericDescriber.opener(
                    **{
                        "describing": "Route in route table {0}".format(
                            self.selection["route table"]
                        ),
                        "content": json.dumps(
                            self.selection.controller_data, sort_keys=True, indent=2
                        ),
                        "pushed": True,
                    }
                )
            )

    def determine_gateway_type(self, entry):
        if "NatGatewayId" in entry:
            return "NAT"
        elif "InstanceId" in entry:
            return "Instance"
        elif "TransitGatewayId" in entry:
            return "Transit"
        elif "LocalGatewayId" in entry:
            return "Local"
        elif "CarrierGatewayId" in entry:
            return "Carrier"
        elif "VpcPeeringConnectionId" in entry:
            return "VPC Peering"
        elif "EgressOnlyInternetGatewayId" in entry:
            return "Egress-Only"
        elif entry["GatewayId"] == "local":
            return "VPC-Local"
        else:
            return "Internet"

    def determine_gateway(self, entry):
        if "NatGatewayId" in entry:
            return entry["NatGatewayId"]
        elif "InstanceId" in entry:
            return entry["InstanceId"]
        elif "TransitGatewayId" in entry:
            return entry["TransitGatewayId"]
        elif "LocalGatewayId" in entry:
            return entry["LocalGatewayId"]
        elif "CarrierGatewayId" in entry:
            return entry["CarrierGatewayId"]
        elif "VpcPeeringConnectionId" in entry:
            return entry["VpcPeeringConnectionId"]
        elif "EgressOnlyInternetGatewayId" in entry:
            return entry["EgressOnlyInternetGatewayId"]
        else:
            return entry["GatewayId"]



import botocore

from .base_control import Describer, ResourceLister
from .common import Common


class R53ResourceLister(ResourceLister):
    prefix = "r53_list"
    title = "Route53 Hosted Zones"
    command_palette = ["r53", "route53"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "route53"
        self.list_method = "list_hosted_zones"
        self.item_path = ".HostedZones"
        self.column_paths = {
            "id": ".Id",
            "name": ".Name",
            "records": ".ResourceRecordSetCount",
        }
        self.imported_column_sizes = {
            "id": 30,
            "name": 30,
            "records": 5,
        }
        self.describe_command = R53Describer.opener
        self.open_command = R53RecordLister.opener
        self.open_selection_arg = "r53_entry"
        self.primary_key = "id"

        self.imported_column_order = ["id", "name", "records"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)


class R53RecordLister(ResourceLister):
    prefix = "r53_record_list"
    title = "Route53 Records"

    def title_info(self):
        return self.r53_entry["name"]

    def __init__(self, *args, r53_entry, **kwargs):
        self.r53_entry = r53_entry
        self.resource_key = "route53"
        self.list_method = "list_resource_record_sets"
        self.list_kwargs = {"HostedZoneId": self.r53_entry["id"]}
        self.item_path = ".ResourceRecordSets"
        self.column_paths = {
            "entry": ".Type",
            "name": self.determine_name,
            "records": self.determine_records,
            "ttl": ".TTL",
        }
        self.hidden_columns = {
            "hosted_zone_id": self.determine_hosted_zone_id,
        }
        self.imported_column_sizes = {
            "entry": 5,
            "name": 30,
            "records": 60,
            "ttl": 5,
        }
        self.describe_command = R53RecordDescriber.opener

        self.imported_column_order = ["entry", "name", "records", "ttl"]
        self.sort_column = "name"
        self.primary_key = None
        super().__init__(*args, **kwargs)
        self.add_hotkey("e", self.edit, "Edit")

    def determine_hosted_zone_id(self, s):
        return self.r53_entry["id"]

    def determine_name(self, s):
        return s["Name"].replace("\\052", "*")

    def edit(self, _):
        if self.selection is not None:
            raw = self.selection.controller_data
            if not self.is_alias(raw):
                content = "\n".join(
                    record["Value"] for record in raw["ResourceRecords"]
                )
                newcontent = Common.Session.textedit(content).strip(" \n\t")
                if content == newcontent:
                    Common.Session.set_message(
                        "Input unchanged.", Common.color("message_info")
                    )
                    return
                newcontent = newcontent.split("\n")
                Common.generic_api_call(
                    "route53",
                    "change_resource_record_sets",
                    {
                        "HostedZoneId": self.selection["hosted_zone_id"],
                        "ChangeBatch": {
                            "Changes": [
                                {
                                    "Action": "UPSERT",
                                    "ResourceRecordSet": {
                                        "Name": self.selection["name"],
                                        "Type": self.selection["entry"],
                                        "ResourceRecords": [
                                            {"Value": line} for line in newcontent
                                        ],
                                        "TTL": int(self.selection["ttl"]),
                                    },
                                }
                            ]
                        },
                    },
                    "Edit DNS record",
                    "Route53",
                    subcategory="Hosted Zone",
                    success_template="Modified DNS entry for zone {0}",
                    resource=self.selection["hosted_zone_id"],
                )

                self.refresh_data()
            else:
                Common.Session.set_message(
                    "Cannot edit aliased records", Common.color("message_info")
                )

    def is_alias(self, s):
        return "AliasTarget" in s and s["AliasTarget"]["DNSName"] != ""

    def determine_records(self, s):
        if self.is_alias(s):
            return s["AliasTarget"]["DNSName"]
        return ",".join([record["Value"] for record in s["ResourceRecords"]])


class R53Describer(Describer):
    prefix = "r53_browser"
    title = "Route53 Hosted Zone"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "route53"
        self.describe_method = "get_hosted_zone"
        self.describe_kwarg_name = "Id"
        self.object_path = "."
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class R53RecordDescriber(Describer):
    prefix = "r53_record_browser"
    title = "Route53 Record"

    def populate_entry(self, *args, entry, **kwargs):
        super().populate_entry(*args, entry=entry, **kwargs)
        self.record_type = entry["entry"]
        self.record_name = entry["name"]

    def populate_describe_kwargs(self, *args, **kwargs):
        super().populate_describe_kwargs(*args, **kwargs)
        self.describe_kwargs["StartRecordType"] = self.record_type
        self.describe_kwargs["StartRecordName"] = self.record_name

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="hosted_zone_id",
        **kwargs
    ):
        self.resource_key = "route53"
        self.describe_method = "list_resource_record_sets"
        self.describe_kwarg_name = "HostedZoneId"
        self.object_path = ".ResourceRecordSets[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )

    def title_info(self):
        return "{0} {1}".format(self.record_type, self.record_name)



class Commons:
    UIInstance = None
    TextfieldInputs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890/:@#.$=_-[]{}()<>|%!+~? "
    TextfieldInputsAlphaNum = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890"
    )

    @classmethod
    def alphanum_with_symbols(cls, symbols):
        return "{0}{1}".format(cls.TextfieldInputsAlphaNum, symbols)
